@@ -11,7 +11,8 @@
 #include "mozilla/Preferences.h"
 #include "mozilla/Services.h"
 #include "mozilla/StaticPrefs_privacy.h"
-#include "mozilla/intl/MozLocale.h"
+#include "mozilla/intl/AppDateTimeFormat.h"
+#include "mozilla/intl/Locale.h"
 #include "mozilla/intl/OSPreferences.h"
 #include "nsDirectoryService.h"
 #include "nsDirectoryServiceDefs.h"
@@ -19,14 +20,22 @@
 #include "nsStringEnumerator.h"
 #include "nsXULAppAPI.h"
 #include "nsZipArchive.h"
+#ifdef XP_WIN
+#  include "WinUtils.h"
+#endif
+#ifdef MOZ_WIDGET_GTK
+#  include "mozilla/WidgetUtilsGtk.h"
+#endif
 
 #define INTL_SYSTEM_LOCALES_CHANGED "intl:system-locales-changed"
 
+#define PSEUDO_LOCALE_PREF "intl.l10n.pseudo"
 #define REQUESTED_LOCALES_PREF "intl.locale.requested"
 #define WEB_EXPOSED_LOCALES_PREF "intl.locale.privacy.web_exposed"
 
 static const char* kObservedPrefs[] = {REQUESTED_LOCALES_PREF,
-                                       WEB_EXPOSED_LOCALES_PREF, nullptr};
+                                       WEB_EXPOSED_LOCALES_PREF,
+                                       PSEUDO_LOCALE_PREF, nullptr};
 
 using namespace mozilla::intl::ffi;
 using namespace mozilla::intl;
@@ -58,18 +67,29 @@ static void SplitLocaleListStringIntoArray(nsACString& str,
 static void ReadRequestedLocales(nsTArray<nsCString>& aRetVal) {
   nsAutoCString str;
   nsresult rv = Preferences::GetCString(REQUESTED_LOCALES_PREF, str);
+  // isRepack means this is a version of Firefox specifically
+  // built for one language.
+  const bool isRepack =
+#ifdef XP_WIN
+      !mozilla::widget::WinUtils::HasPackageIdentity();
+#elif defined(MOZ_WIDGET_GTK)
+      !widget::IsRunningUnderSnap();
+#else
+      true;
+#endif
 
-  // We handle three scenarios here:
+  // We handle four scenarios here:
   //
   // 1) The pref is not set - use default locale
-  // 2) The pref is set to "" - use OS locales
-  // 3) The pref is set to a value - parse the locale list and use it
+  // 2) The pref is not set and we're a packaged app - use OS locales
+  // 3) The pref is set to "" - use OS locales
+  // 4) The pref is set to a value - parse the locale list and use it
   if (NS_SUCCEEDED(rv)) {
     if (str.Length() == 0) {
-      // If the pref string is empty, we'll take requested locales
-      // from the OS.
+      // Case 3
       OSPreferences::GetInstance()->GetSystemLocales(aRetVal);
     } else {
+      // Case 4
       SplitLocaleListStringIntoArray(str, aRetVal);
     }
   }
@@ -78,9 +98,15 @@ static void ReadRequestedLocales(nsTArray<nsCString>& aRetVal) {
   // or parsing of the pref didn't produce any usable
   // result.
   if (aRetVal.IsEmpty()) {
-    nsAutoCString defaultLocale;
-    LocaleService::GetInstance()->GetDefaultLocale(defaultLocale);
-    aRetVal.AppendElement(defaultLocale);
+    if (isRepack) {
+      // Case 1
+      nsAutoCString defaultLocale;
+      LocaleService::GetInstance()->GetDefaultLocale(defaultLocale);
+      aRetVal.AppendElement(defaultLocale);
+    } else {
+      // Case 2
+      OSPreferences::GetInstance()->GetSystemLocales(aRetVal);
+    }
   }
 }
 
@@ -157,6 +183,15 @@ LocaleService* LocaleService::GetInstance() {
   return sInstance;
 }
 
+static void NotifyAppLocaleChanged() {
+  nsCOMPtr<nsIObserverService> obs = mozilla::services::GetObserverService();
+  if (obs) {
+    obs->NotifyObservers(nullptr, "intl:app-locales-changed", nullptr);
+  }
+  // The locale in AppDateTimeFormat is cached statically.
+  AppDateTimeFormat::ClearLocaleCache();
+}
+
 void LocaleService::RemoveObservers() {
   if (mIsServer) {
     Preferences::RemoveObservers(this, kObservedPrefs);
@@ -174,10 +209,7 @@ void LocaleService::AssignAppLocales(const nsTArray<nsCString>& aAppLocales) {
              "This should only be called for LocaleService in client mode.");
 
   mAppLocales = aAppLocales.Clone();
-  nsCOMPtr<nsIObserverService> obs = mozilla::services::GetObserverService();
-  if (obs) {
-    obs->NotifyObservers(nullptr, "intl:app-locales-changed", nullptr);
-  }
+  NotifyAppLocaleChanged();
 }
 
 void LocaleService::AssignRequestedLocales(
@@ -231,10 +263,7 @@ void LocaleService::LocalesChanged() {
 
   if (mAppLocales != newLocales) {
     mAppLocales = std::move(newLocales);
-    nsCOMPtr<nsIObserverService> obs = mozilla::services::GetObserverService();
-    if (obs) {
-      obs->NotifyObservers(nullptr, "intl:app-locales-changed", nullptr);
-    }
+    NotifyAppLocaleChanged();
   }
 }
 
@@ -277,6 +306,8 @@ LocaleService::Observe(nsISupports* aSubject, const char* aTopic,
       RequestedLocalesChanged();
     } else if (pref.EqualsLiteral(WEB_EXPOSED_LOCALES_PREF)) {
       WebExposedLocalesChanged();
+    } else if (pref.EqualsLiteral(PSEUDO_LOCALE_PREF)) {
+      NotifyAppLocaleChanged();
     }
   }
 
@@ -285,9 +316,20 @@ LocaleService::Observe(nsISupports* aSubject, const char* aTopic,
 
 bool LocaleService::LanguagesMatch(const nsACString& aRequested,
                                    const nsACString& aAvailable) {
-  Locale requested = Locale(aRequested);
-  Locale available = Locale(aAvailable);
-  return requested.GetLanguage().Equals(available.GetLanguage());
+  Locale requested;
+  auto requestedResult = LocaleParser::TryParse(aRequested, requested);
+  Locale available;
+  auto availableResult = LocaleParser::TryParse(aAvailable, available);
+
+  if (requestedResult.isErr() || availableResult.isErr()) {
+    return false;
+  }
+
+  if (requested.Canonicalize().isErr() || available.Canonicalize().isErr()) {
+    return false;
+  }
+
+  return requested.Language().Span() == available.Language().Span();
 }
 
 bool LocaleService::IsServer() { return mIsServer; }
@@ -296,7 +338,7 @@ static bool GetGREFileContents(const char* aFilePath, nsCString* aOutString) {
   // Look for the requested file in omnijar.
   RefPtr<nsZipArchive> zip = Omnijar::GetReader(Omnijar::GRE);
   if (zip) {
-    nsZipItemPtr<char> item(zip, aFilePath);
+    nsZipItemPtr<char> item(zip, nsDependentCString(aFilePath));
     if (!item) {
       return false;
     }
@@ -509,9 +551,14 @@ LocaleService::NegotiateLanguages(const nsTArray<nsCString>& aRequested,
     return NS_ERROR_INVALID_ARG;
   }
 
+#ifdef DEBUG
+  Locale parsedLocale;
+  auto result = LocaleParser::TryParse(aDefaultLocale, parsedLocale);
+
   MOZ_ASSERT(
-      aDefaultLocale.IsEmpty() || Locale(aDefaultLocale).IsWellFormed(),
+      aDefaultLocale.IsEmpty() || result.isOk(),
       "If specified, default locale must be a well-formed BCP47 language tag.");
+#endif
 
   if (aStrategy == kLangNegStrategyLookup && aDefaultLocale.IsEmpty()) {
     NS_WARNING(
