@@ -18,11 +18,15 @@ use crate::values::generics::position::PositionComponent as GenericPositionCompo
 use crate::values::generics::position::PositionOrAuto as GenericPositionOrAuto;
 use crate::values::generics::position::ZIndex as GenericZIndex;
 use crate::values::specified::{AllowQuirks, Integer, LengthPercentage, NonNegativeNumber};
+use crate::values::DashedIdent;
 use crate::{Atom, Zero};
 use cssparser::Parser;
 use selectors::parser::SelectorParseErrorKind;
 use servo_arc::Arc;
+use smallvec::{smallvec, SmallVec};
+use std::collections::hash_map::Entry;
 use std::fmt::{self, Write};
+use style_traits::arc_slice::ArcSlice;
 use style_traits::values::specified::AllowedNumericType;
 use style_traits::{CssWriter, ParseError, StyleParseErrorKind, ToCss};
 
@@ -224,12 +228,12 @@ impl ToCss for Position {
             ) => {
                 dest.write_str("left ")?;
                 x_lp.to_css(dest)?;
-                dest.write_str(" ")?;
+                dest.write_char(' ')?;
                 y_pos.to_css(dest)
             },
             (x_pos, y_pos) => {
                 x_pos.to_css(dest)?;
-                dest.write_str(" ")?;
+                dest.write_char(' ')?;
                 y_pos.to_css(dest)
             },
         }
@@ -336,6 +340,762 @@ impl<S: Side> PositionComponent<S> {
     }
 }
 
+/// https://drafts.csswg.org/css-anchor-position-1/#propdef-anchor-name
+#[repr(transparent)]
+#[derive(
+    Clone,
+    Debug,
+    MallocSizeOf,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[css(comma)]
+pub struct AnchorName(
+    #[css(iterable, if_empty = "none")]
+    #[ignore_malloc_size_of = "Arc"]
+    pub crate::ArcSlice<DashedIdent>,
+);
+
+impl AnchorName {
+    /// Return the `none` value.
+    pub fn none() -> Self {
+        Self(Default::default())
+    }
+
+    /// Returns whether this is the `none` value.
+    pub fn is_none(&self) -> bool {
+        self.0.is_empty()
+    }
+}
+
+impl Parse for AnchorName {
+    fn parse<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        let location = input.current_source_location();
+        let first = input.expect_ident()?;
+        if first.eq_ignore_ascii_case("none") {
+            return Ok(Self::none());
+        }
+        // The common case is probably just to have a single anchor name, so
+        // space for four on the stack should be plenty.
+        let mut idents: SmallVec<[DashedIdent; 4]> =
+            smallvec![DashedIdent::from_ident(location, first,)?];
+        while input.try_parse(|input| input.expect_comma()).is_ok() {
+            idents.push(DashedIdent::parse(context, input)?);
+        }
+        Ok(AnchorName(ArcSlice::from_iter(idents.drain(..))))
+    }
+}
+
+/// https://drafts.csswg.org/css-anchor-position-1/#propdef-scope
+#[derive(
+    Clone,
+    Debug,
+    MallocSizeOf,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[repr(u8)]
+pub enum AnchorScope {
+    /// `none`
+    None,
+    /// `all`
+    All,
+    /// `<dashed-ident>#`
+    #[css(comma)]
+    Idents(
+        #[css(iterable)]
+        #[ignore_malloc_size_of = "Arc"]
+        crate::ArcSlice<DashedIdent>,
+    ),
+}
+
+impl AnchorScope {
+    /// Return the `none` value.
+    pub fn none() -> Self {
+        Self::None
+    }
+
+    /// Returns whether this is the `none` value.
+    pub fn is_none(&self) -> bool {
+        *self == Self::None
+    }
+}
+
+impl Parse for AnchorScope {
+    fn parse<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        let location = input.current_source_location();
+        let first = input.expect_ident()?;
+        if first.eq_ignore_ascii_case("none") {
+            return Ok(Self::None);
+        }
+        if first.eq_ignore_ascii_case("all") {
+            return Ok(Self::All);
+        }
+        // Authors using more than a handful of anchored elements is likely
+        // uncommon, so we only pre-allocate for 8 on the stack here.
+        let mut idents: SmallVec<[DashedIdent; 8]> =
+            smallvec![DashedIdent::from_ident(location, first,)?];
+        while input.try_parse(|input| input.expect_comma()).is_ok() {
+            idents.push(DashedIdent::parse(context, input)?);
+        }
+        Ok(AnchorScope::Idents(ArcSlice::from_iter(idents.drain(..))))
+    }
+}
+
+/// https://drafts.csswg.org/css-anchor-position-1/#propdef-position-anchor
+#[derive(
+    Clone,
+    Debug,
+    MallocSizeOf,
+    Parse,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[repr(u8)]
+pub enum PositionAnchor {
+    /// `auto`
+    Auto,
+    /// `<dashed-ident>`
+    Ident(DashedIdent),
+}
+
+impl PositionAnchor {
+    /// Return the `auto` value.
+    pub fn auto() -> Self {
+        Self::Auto
+    }
+
+    /// Returns whether this is the `auto` value.
+    pub fn is_auto(&self) -> bool {
+        *self == Self::Auto
+    }
+}
+
+#[derive(
+    Clone,
+    Copy,
+    Debug,
+    Eq,
+    MallocSizeOf,
+    Parse,
+    PartialEq,
+    Serialize,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[css(bitflags(mixed = "flip-block,flip-inline,flip-start"))]
+#[repr(C)]
+/// https://drafts.csswg.org/css-anchor-position-1/#typedef-position-try-options-try-tactic
+/// <try-tactic>
+pub struct PositionTryOptionsTryTactic(u8);
+bitflags! {
+    impl PositionTryOptionsTryTactic: u8 {
+        /// `flip-block`
+        const FLIP_BLOCK = 1 << 0;
+        /// `flip-inline`
+        const FLIP_INLINE = 1 << 1;
+        /// `flip-start`
+        const FLIP_START = 1 << 2;
+    }
+}
+
+#[derive(
+    Clone,
+    Debug,
+    MallocSizeOf,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[repr(C)]
+/// https://drafts.csswg.org/css-anchor-position-1/#propdef-position-try-options
+/// <dashed-ident> || <try-tactic>
+pub struct DashedIdentAndOrTryTactic {
+    /// `<dashed-ident>`
+    pub ident: DashedIdent,
+    /// `<try-tactic>`
+    pub try_tactic: PositionTryOptionsTryTactic,
+}
+
+impl Parse for DashedIdentAndOrTryTactic {
+    fn parse<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        let mut result = Self {
+            ident: DashedIdent::empty(),
+            try_tactic: PositionTryOptionsTryTactic::empty(),
+        };
+
+        loop {
+            if result.ident.is_empty() {
+                if let Ok(ident) = input.try_parse(|i| DashedIdent::parse(context, i)) {
+                    result.ident = ident;
+                    continue;
+                }
+            }
+            if result.try_tactic.is_empty() {
+                if let Ok(try_tactic) =
+                    input.try_parse(|i| PositionTryOptionsTryTactic::parse(context, i))
+                {
+                    result.try_tactic = try_tactic;
+                    continue;
+                }
+            }
+            break;
+        }
+
+        if result.ident.is_empty() && result.try_tactic.is_empty() {
+            return Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError));
+        }
+        return Ok(result);
+    }
+}
+
+#[derive(
+    Clone,
+    Debug,
+    MallocSizeOf,
+    Parse,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[repr(u8)]
+/// https://drafts.csswg.org/css-anchor-position-1/#propdef-position-try-options
+/// [ [<dashed-ident> || <try-tactic>] | <'inset-area'> ]
+pub enum PositionTryOptionsItem {
+    /// `<dashed-ident> || <try-tactic>`
+    IdentAndOrTactic(DashedIdentAndOrTryTactic),
+    #[parse(parse_fn = "InsetArea::parse_except_none")]
+    /// `<inset-area>`
+    InsetArea(InsetArea),
+}
+
+#[derive(
+    Clone,
+    Debug,
+    Default,
+    MallocSizeOf,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[css(comma)]
+#[repr(C)]
+/// https://drafts.csswg.org/css-anchor-position-1/#position-try-options
+pub struct PositionTryOptions(
+    #[css(iterable, if_empty = "none")]
+    #[ignore_malloc_size_of = "Arc"]
+    pub crate::ArcSlice<PositionTryOptionsItem>,
+);
+
+impl PositionTryOptions {
+    #[inline]
+    /// Return the `none` value.
+    pub fn none() -> Self {
+        Self(Default::default())
+    }
+
+    /// Returns whether this is the `none` value.
+    pub fn is_none(&self) -> bool {
+        self.0.is_empty()
+    }
+}
+
+impl Parse for PositionTryOptions {
+    fn parse<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        if input.try_parse(|i| i.expect_ident_matching("none")).is_ok() {
+            return Ok(Self::none());
+        }
+        // The common case is unlikely to include many alternate positioning
+        // styles, so space for four on the stack should typically be enough.
+        let mut items: SmallVec<[PositionTryOptionsItem; 4]> =
+            smallvec![PositionTryOptionsItem::parse(context, input)?];
+        while input.try_parse(|input| input.expect_comma()).is_ok() {
+            items.push(PositionTryOptionsItem::parse(context, input)?);
+        }
+        Ok(Self(ArcSlice::from_iter(items.drain(..))))
+    }
+}
+
+/// https://drafts.csswg.org/css-anchor-position-1/#position-try-order-property
+#[derive(
+    Clone,
+    Copy,
+    Debug,
+    Default,
+    Eq,
+    MallocSizeOf,
+    Parse,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[repr(u8)]
+pub enum PositionTryOrder {
+    #[default]
+    /// `normal`
+    Normal,
+    /// `most-width`
+    MostWidth,
+    /// `most-height`
+    MostHeight,
+    /// `most-block-size`
+    MostBlockSize,
+    /// `most-inline-size`
+    MostInlineSize,
+}
+
+impl PositionTryOrder {
+    #[inline]
+    /// Return the `auto` value.
+    pub fn normal() -> Self {
+        Self::Normal
+    }
+
+    /// Returns whether this is the `auto` value.
+    pub fn is_normal(&self) -> bool {
+        *self == Self::Normal
+    }
+}
+
+#[derive(
+    Clone,
+    Copy,
+    Debug,
+    Eq,
+    MallocSizeOf,
+    Parse,
+    PartialEq,
+    Serialize,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[css(bitflags(single = "always", mixed = "anchors-valid,anchors-visible,no-overflow"))]
+#[repr(C)]
+/// Specified keyword values for the position-visibility property.
+pub struct PositionVisibility(u8);
+bitflags! {
+    impl PositionVisibility: u8 {
+        /// Element is displayed without regard for its anchors or its overflowing status.
+        const ALWAYS = 0;
+        /// anchors-valid
+        const ANCHORS_VALID = 1 << 0;
+        /// anchors-visible
+        const ANCHORS_VISIBLE = 1 << 1;
+        /// no-overflow
+        const NO_OVERFLOW = 1 << 2;
+    }
+}
+
+impl Default for PositionVisibility {
+    fn default() -> Self {
+        Self::ALWAYS
+    }
+}
+
+impl PositionVisibility {
+    #[inline]
+    /// Returns the initial value of position-visibility
+    pub fn always() -> Self {
+        Self::ALWAYS
+    }
+}
+
+#[derive(
+    Clone,
+    Copy,
+    Debug,
+    Default,
+    Eq,
+    MallocSizeOf,
+    Parse,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[allow(missing_docs)]
+#[repr(u8)]
+/// Possible values for the `inset-area` preperty's keywords.
+/// https://drafts.csswg.org/css-anchor-position-1/#propdef-inset-area
+pub enum InsetAreaKeyword {
+    #[default]
+    None,
+
+    // Common (shared) keywords:
+    Center,
+    SpanAll,
+
+    // Horizontal keywords:
+    Left,
+    Right,
+    SpanLeft,
+    SpanRight,
+    XStart,
+    XEnd,
+    SpanXStart,
+    SpanXEnd,
+    XSelfStart,
+    XSelfEnd,
+    SpanXSelfStart,
+    SpanXSelfEnd,
+    // Vertical keywords:
+    Top,
+    Bottom,
+    SpanTop,
+    SpanBottom,
+    YStart,
+    YEnd,
+    SpanYStart,
+    SpanYEnd,
+    YSelfStart,
+    YSelfEnd,
+    SpanYSelfStart,
+    SpanYSelfEnd,
+
+    // Block keywords:
+    BlockStart,
+    BlockEnd,
+    SpanBlockStart,
+    SpanBlockEnd,
+    // Inline keywords:
+    InlineStart,
+    InlineEnd,
+    SpanInlineStart,
+    SpanInlineEnd,
+
+    // "Self" block keywords:
+    SelfBlockStart,
+    SelfBlockEnd,
+    SpanSelfBlockStart,
+    SpanSelfBlockEnd,
+    // "Self" inline keywords:
+    SelfInlineStart,
+    SelfInlineEnd,
+    SpanSelfInlineStart,
+    SpanSelfInlineEnd,
+
+    // Inferred axis keywords:
+    Start,
+    End,
+    SpanStart,
+    SpanEnd,
+
+    // "Self" inferred axis keywords:
+    SelfStart,
+    SelfEnd,
+    SpanSelfStart,
+    SpanSelfEnd,
+}
+
+#[allow(missing_docs)]
+impl InsetAreaKeyword {
+    #[inline]
+    pub fn none() -> Self {
+        Self::None
+    }
+
+    pub fn is_none(&self) -> bool {
+        *self == Self::None
+    }
+
+    /// Is a value that's common to all compatible keyword groupings.
+    pub fn is_common(&self) -> bool {
+        *self == Self::Center || *self == Self::SpanAll
+    }
+
+    pub fn is_horizontal(&self) -> bool {
+        matches!(
+            self,
+            Self::Left |
+                Self::Right |
+                Self::SpanLeft |
+                Self::SpanRight |
+                Self::XStart |
+                Self::XEnd |
+                Self::SpanXStart |
+                Self::SpanXEnd |
+                Self::XSelfStart |
+                Self::XSelfEnd |
+                Self::SpanXSelfStart |
+                Self::SpanXSelfEnd
+        )
+    }
+    pub fn is_vertical(&self) -> bool {
+        matches!(
+            self,
+            Self::Top |
+                Self::Bottom |
+                Self::SpanTop |
+                Self::SpanBottom |
+                Self::YStart |
+                Self::YEnd |
+                Self::SpanYStart |
+                Self::SpanYEnd |
+                Self::YSelfStart |
+                Self::YSelfEnd |
+                Self::SpanYSelfStart |
+                Self::SpanYSelfEnd
+        )
+    }
+
+    pub fn is_block(&self) -> bool {
+        matches!(
+            self,
+            Self::BlockStart | Self::BlockEnd | Self::SpanBlockStart | Self::SpanBlockEnd
+        )
+    }
+    pub fn is_inline(&self) -> bool {
+        matches!(
+            self,
+            Self::InlineStart | Self::InlineEnd | Self::SpanInlineStart | Self::SpanInlineEnd
+        )
+    }
+
+    pub fn is_self_block(&self) -> bool {
+        matches!(
+            self,
+            Self::SelfBlockStart |
+                Self::SelfBlockEnd |
+                Self::SpanSelfBlockStart |
+                Self::SpanSelfBlockEnd
+        )
+    }
+    pub fn is_self_inline(&self) -> bool {
+        matches!(
+            self,
+            Self::SelfInlineStart |
+                Self::SelfInlineEnd |
+                Self::SpanSelfInlineStart |
+                Self::SpanSelfInlineEnd
+        )
+    }
+
+    pub fn is_inferred_logical(&self) -> bool {
+        matches!(
+            self,
+            Self::Start | Self::End | Self::SpanStart | Self::SpanEnd
+        )
+    }
+
+    pub fn is_self_inferred_logical(&self) -> bool {
+        matches!(
+            self,
+            Self::SelfStart | Self::SelfEnd | Self::SpanSelfStart | Self::SpanSelfEnd
+        )
+    }
+}
+
+#[inline]
+fn is_compatible_pairing(first: InsetAreaKeyword, second: InsetAreaKeyword) -> bool {
+    if first.is_none() || second.is_none() {
+        // `none` is not allowed as one of the keywords when two keywords are
+        // provided.
+        return false;
+    }
+    if first.is_common() || second.is_common() {
+        return true;
+    }
+    if first.is_horizontal() {
+        return second.is_vertical();
+    }
+    if first.is_vertical() {
+        return second.is_horizontal();
+    }
+    if first.is_block() {
+        return second.is_inline();
+    }
+    if first.is_inline() {
+        return second.is_block();
+    }
+    if first.is_self_block() {
+        return second.is_self_inline();
+    }
+    if first.is_self_inline() {
+        return second.is_self_block();
+    }
+    if first.is_inferred_logical() {
+        return second.is_inferred_logical();
+    }
+    if first.is_self_inferred_logical() {
+        return second.is_self_inferred_logical();
+    }
+
+    debug_assert!(false, "Not reached");
+
+    // Return false to increase the chances of this being reported to us if we
+    // ever were to get here.
+    false
+}
+
+#[derive(
+    Clone,
+    Copy,
+    Debug,
+    Eq,
+    MallocSizeOf,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToCss,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[repr(C)]
+/// https://drafts.csswg.org/css-anchor-position-1/#propdef-inset-area
+pub struct InsetArea {
+    /// First keyword, if any.
+    pub first: InsetAreaKeyword,
+    /// Second keyword, if any.
+    #[css(skip_if = "InsetAreaKeyword::is_none")]
+    pub second: InsetAreaKeyword,
+}
+
+#[allow(missing_docs)]
+impl InsetArea {
+    #[inline]
+    pub fn none() -> Self {
+        Self {
+            first: InsetAreaKeyword::None,
+            second: InsetAreaKeyword::None,
+        }
+    }
+
+    #[inline]
+    pub fn is_none(&self) -> bool {
+        self.first.is_none()
+    }
+
+    pub fn parse_except_none<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        Self::parse_internal(context, input, /*allow_none*/ false)
+    }
+
+    fn parse_internal<'i, 't>(
+        _context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+        allow_none: bool,
+    ) -> Result<Self, ParseError<'i>> {
+        let mut location = input.current_source_location();
+        let mut first = InsetAreaKeyword::parse(input)?;
+        if first.is_none() {
+            if allow_none {
+                return Ok(Self::none());
+            }
+            return Err(location.new_custom_error(StyleParseErrorKind::UnspecifiedError));
+        }
+
+        location = input.current_source_location();
+        let second = input.try_parse(InsetAreaKeyword::parse);
+        if let Ok(InsetAreaKeyword::None) = second {
+            // `none` is only allowed as a single value
+            return Err(location.new_custom_error(StyleParseErrorKind::UnspecifiedError));
+        }
+        let mut second = second.unwrap_or(InsetAreaKeyword::None);
+        if second.is_none() {
+            // Either there was no second keyword and try_parse returned a
+            // BasicParseErrorKind::EndOfInput, or else the second "keyword"
+            // was invalid. We assume the former case here, and if it's the
+            // latter case then our caller detects the error (try_parse will,
+            // have rewound, leaving an unparsed token).
+            return Ok(Self { first, second });
+        }
+
+        if !is_compatible_pairing(first, second) {
+            return Err(location.new_custom_error(StyleParseErrorKind::UnspecifiedError));
+        }
+
+        // Normalize by applying the shortest serialization principle:
+        // https://drafts.csswg.org/cssom/#serializing-css-values
+        if first.is_inferred_logical() ||
+            second.is_inferred_logical() ||
+            first.is_self_inferred_logical() ||
+            second.is_self_inferred_logical() ||
+            (first.is_common() && second.is_common())
+        {
+            // In these cases we must not change the order of the keywords
+            // since their meaning is inferred from their order. However, if
+            // both keywords are the same, only one should be set.
+            if first == second {
+                second = InsetAreaKeyword::None;
+            }
+        } else if second == InsetAreaKeyword::SpanAll {
+            // Span-all is the default behavior, so specifying `span-all` is
+            // superfluous.
+            second = InsetAreaKeyword::None;
+        } else if first == InsetAreaKeyword::SpanAll {
+            // Same here, but the non-superfluous keyword must come first.
+            first = second;
+            second = InsetAreaKeyword::None;
+        } else if first.is_vertical() ||
+            second.is_horizontal() ||
+            first.is_inline() ||
+            second.is_block() ||
+            first.is_self_inline() ||
+            second.is_self_block()
+        {
+            // Canonical order is horizontal before vertical, block before inline.
+            std::mem::swap(&mut first, &mut second);
+        }
+
+        Ok(Self { first, second })
+    }
+}
+
+impl Parse for InsetArea {
+    fn parse<'i, 't>(
+        context: &ParserContext,
+        input: &mut Parser<'i, 't>,
+    ) -> Result<Self, ParseError<'i>> {
+        Self::parse_internal(context, input, /*allow_none*/ true)
+    }
+}
+
 /// Represents a side, either horizontal or vertical, of a CSS position.
 pub trait Side {
     /// Returns the start side.
@@ -369,19 +1129,30 @@ impl Side for VerticalPositionKeyword {
     }
 }
 
+/// Controls how the auto-placement algorithm works specifying exactly how auto-placed items
+/// get flowed into the grid: [ row | column ] || dense
+/// https://drafts.csswg.org/css-grid-2/#grid-auto-flow-property
+#[derive(
+    Clone,
+    Copy,
+    Debug,
+    Eq,
+    MallocSizeOf,
+    Parse,
+    PartialEq,
+    SpecifiedValueInfo,
+    ToComputedValue,
+    ToResolvedValue,
+    ToShmem,
+)]
+#[css(bitflags(
+    mixed = "row,column,dense",
+    validate_mixed = "Self::validate_and_simplify"
+))]
+#[repr(C)]
+pub struct GridAutoFlow(u8);
 bitflags! {
-    /// Controls how the auto-placement algorithm works
-    /// specifying exactly how auto-placed items get flowed into the grid
-    #[derive(
-        MallocSizeOf,
-        SpecifiedValueInfo,
-        ToComputedValue,
-        ToResolvedValue,
-        ToShmem
-    )]
-    #[value_info(other_values = "row,column,dense")]
-    #[repr(C)]
-    pub struct GridAutoFlow: u8 {
+    impl GridAutoFlow: u8 {
         /// 'row' - mutually exclusive with 'column'
         const ROW = 1 << 0;
         /// 'column' - mutually exclusive with 'row'
@@ -391,6 +1162,44 @@ bitflags! {
     }
 }
 
+impl GridAutoFlow {
+    /// [ row | column ] || dense
+    fn validate_and_simplify(&mut self) -> bool {
+        if self.contains(Self::ROW | Self::COLUMN) {
+            // row and column are mutually exclusive.
+            return false;
+        }
+        if *self == Self::DENSE {
+            // If there's no column, default to row.
+            self.insert(Self::ROW);
+        }
+        true
+    }
+}
+
+impl ToCss for GridAutoFlow {
+    fn to_css<W>(&self, dest: &mut CssWriter<W>) -> fmt::Result
+    where
+        W: Write,
+    {
+        let dense = self.intersects(Self::DENSE);
+        if self.intersects(Self::ROW) {
+            return if dense {
+                dest.write_str("dense")
+            } else {
+                dest.write_str("row")
+            };
+        }
+        debug_assert!(self.intersects(Self::COLUMN));
+        if dense {
+            dest.write_str("column dense")
+        } else {
+            dest.write_str("column")
+        }
+    }
+}
+
+#[repr(u8)]
 #[derive(
     Clone,
     Copy,
@@ -412,6 +1221,7 @@ pub enum MasonryPlacement {
     Next,
 }
 
+#[repr(u8)]
 #[derive(
     Clone,
     Copy,
@@ -446,6 +1256,7 @@ pub enum MasonryItemOrder {
     ToResolvedValue,
     ToShmem,
 )]
+#[repr(C)]
 /// Controls how the Masonry layout algorithm works
 /// specifying exactly how auto-placed items get flowed in the masonry axis.
 pub struct MasonryAutoFlow {
@@ -525,104 +1336,6 @@ impl Parse for MasonryAutoFlow {
     }
 }
 
-#[cfg(feature = "gecko")]
-impl From<u8> for MasonryAutoFlow {
-    fn from(bits: u8) -> MasonryAutoFlow {
-        use crate::gecko_bindings::structs;
-        let mut value = MasonryAutoFlow::initial();
-        if bits & structs::NS_STYLE_MASONRY_PLACEMENT_PACK as u8 == 0 {
-            value.placement = MasonryPlacement::Next;
-        }
-        if bits & structs::NS_STYLE_MASONRY_ORDER_DEFINITE_FIRST as u8 == 0 {
-            value.order = MasonryItemOrder::Ordered;
-        }
-        value
-    }
-}
-
-#[cfg(feature = "gecko")]
-impl From<MasonryAutoFlow> for u8 {
-    fn from(v: MasonryAutoFlow) -> u8 {
-        use crate::gecko_bindings::structs;
-
-        let mut result: u8 = 0;
-        if v.placement == MasonryPlacement::Pack {
-            result |= structs::NS_STYLE_MASONRY_PLACEMENT_PACK as u8;
-        }
-        if v.order == MasonryItemOrder::DefiniteFirst {
-            result |= structs::NS_STYLE_MASONRY_ORDER_DEFINITE_FIRST as u8;
-        }
-        result
-    }
-}
-
-impl Parse for GridAutoFlow {
-    /// [ row | column ] || dense
-    fn parse<'i, 't>(
-        _context: &ParserContext,
-        input: &mut Parser<'i, 't>,
-    ) -> Result<GridAutoFlow, ParseError<'i>> {
-        let mut track = None;
-        let mut dense = GridAutoFlow::empty();
-
-        while !input.is_exhausted() {
-            let location = input.current_source_location();
-            let ident = input.expect_ident()?;
-            let success = match_ignore_ascii_case! { &ident,
-                "row" if track.is_none() => {
-                    track = Some(GridAutoFlow::ROW);
-                    true
-                },
-                "column" if track.is_none() => {
-                    track = Some(GridAutoFlow::COLUMN);
-                    true
-                },
-                "dense" if dense.is_empty() => {
-                    dense = GridAutoFlow::DENSE;
-                    true
-                },
-                _ => false,
-            };
-            if !success {
-                return Err(location
-                    .new_custom_error(SelectorParseErrorKind::UnexpectedIdent(ident.clone())));
-            }
-        }
-
-        if track.is_some() || !dense.is_empty() {
-            Ok(track.unwrap_or(GridAutoFlow::ROW) | dense)
-        } else {
-            Err(input.new_custom_error(StyleParseErrorKind::UnspecifiedError))
-        }
-    }
-}
-
-impl ToCss for GridAutoFlow {
-    fn to_css<W>(&self, dest: &mut CssWriter<W>) -> fmt::Result
-    where
-        W: Write,
-    {
-        if *self == GridAutoFlow::ROW {
-            return dest.write_str("row");
-        }
-
-        if *self == GridAutoFlow::COLUMN {
-            return dest.write_str("column");
-        }
-
-        if *self == GridAutoFlow::ROW | GridAutoFlow::DENSE {
-            return dest.write_str("dense");
-        }
-
-        if *self == GridAutoFlow::COLUMN | GridAutoFlow::DENSE {
-            return dest.write_str("column dense");
-        }
-
-        debug_assert!(false, "Unknown or invalid grid-autoflow value");
-        Ok(())
-    }
-}
-
 #[derive(
     Clone,
     Debug,
@@ -640,7 +1353,10 @@ pub struct TemplateAreas {
     /// `named area` containing for each template area
     #[css(skip)]
     pub areas: crate::OwnedSlice<NamedArea>,
-    /// The original CSS string value of each template area
+    /// The simplified CSS strings for serialization purpose.
+    /// https://drafts.csswg.org/css-grid/#serialize-template
+    // Note: We also use the length of `strings` when computing the explicit grid end line number
+    // (i.e. row number).
     #[css(iterable)]
     pub strings: crate::OwnedSlice<crate::OwnedStr>,
     /// The number of columns of the grid.
@@ -648,106 +1364,140 @@ pub struct TemplateAreas {
     pub width: u32,
 }
 
-impl TemplateAreas {
-    /// Transform `vector` of str into `template area`
-    pub fn from_vec(strings: Vec<crate::OwnedStr>) -> Result<Self, ()> {
-        if strings.is_empty() {
-            return Err(());
-        }
-        let mut areas: Vec<NamedArea> = vec![];
-        let mut width = 0;
-        {
-            let mut row = 0u32;
-            let mut area_indices = PrecomputedHashMap::<Atom, usize>::default();
-            for string in &strings {
-                let mut current_area_index: Option<usize> = None;
-                row += 1;
-                let mut column = 0u32;
-                for token in TemplateAreasTokenizer(string) {
-                    column += 1;
-                    let name = if let Some(token) = token? {
-                        Atom::from(token)
-                    } else {
-                        if let Some(index) = current_area_index.take() {
-                            if areas[index].columns.end != column {
-                                return Err(());
-                            }
-                        }
-                        continue;
-                    };
-                    if let Some(index) = current_area_index {
-                        if areas[index].name == name {
-                            if areas[index].rows.start == row {
-                                areas[index].columns.end += 1;
-                            }
-                            continue;
-                        }
-                        if areas[index].columns.end != column {
-                            return Err(());
-                        }
+/// Parser for grid template areas.
+#[derive(Default)]
+pub struct TemplateAreasParser {
+    areas: Vec<NamedArea>,
+    area_indices: PrecomputedHashMap<Atom, usize>,
+    strings: Vec<crate::OwnedStr>,
+    width: u32,
+    row: u32,
+}
+
+impl TemplateAreasParser {
+    /// Parse a single string.
+    pub fn try_parse_string<'i>(
+        &mut self,
+        input: &mut Parser<'i, '_>,
+    ) -> Result<(), ParseError<'i>> {
+        input.try_parse(|input| {
+            self.parse_string(input.expect_string()?)
+                .map_err(|()| input.new_custom_error(StyleParseErrorKind::UnspecifiedError))
+        })
+    }
+
+    /// Parse a single string.
+    fn parse_string(&mut self, string: &str) -> Result<(), ()> {
+        self.row += 1;
+        let mut simplified_string = String::new();
+        let mut current_area_index: Option<usize> = None;
+        let mut column = 0u32;
+        for token in TemplateAreasTokenizer(string) {
+            column += 1;
+            if column > 1 {
+                simplified_string.push(' ');
+            }
+            let name = if let Some(token) = token? {
+                simplified_string.push_str(token);
+                Atom::from(token)
+            } else {
+                if let Some(index) = current_area_index.take() {
+                    if self.areas[index].columns.end != column {
+                        return Err(());
                     }
-                    if let Some(index) = area_indices.get(&name).cloned() {
-                        if areas[index].columns.start != column || areas[index].rows.end != row {
-                            return Err(());
-                        }
-                        areas[index].rows.end += 1;
-                        current_area_index = Some(index);
-                        continue;
+                }
+                simplified_string.push('.');
+                continue;
+            };
+            if let Some(index) = current_area_index {
+                if self.areas[index].name == name {
+                    if self.areas[index].rows.start == self.row {
+                        self.areas[index].columns.end += 1;
                     }
-                    let index = areas.len();
-                    assert!(area_indices.insert(name.clone(), index).is_none());
-                    areas.push(NamedArea {
+                    continue;
+                }
+                if self.areas[index].columns.end != column {
+                    return Err(());
+                }
+            }
+            match self.area_indices.entry(name) {
+                Entry::Occupied(ref e) => {
+                    let index = *e.get();
+                    if self.areas[index].columns.start != column ||
+                        self.areas[index].rows.end != self.row
+                    {
+                        return Err(());
+                    }
+                    self.areas[index].rows.end += 1;
+                    current_area_index = Some(index);
+                },
+                Entry::Vacant(v) => {
+                    let index = self.areas.len();
+                    let name = v.key().clone();
+                    v.insert(index);
+                    self.areas.push(NamedArea {
                         name,
                         columns: UnsignedRange {
                             start: column,
                             end: column + 1,
                         },
                         rows: UnsignedRange {
-                            start: row,
-                            end: row + 1,
+                            start: self.row,
+                            end: self.row + 1,
                         },
                     });
                     current_area_index = Some(index);
-                }
-                if column == 0 {
-                    // Each string must produce a valid token.
-                    // https://github.com/w3c/csswg-drafts/issues/5110
-                    return Err(());
-                }
-                if let Some(index) = current_area_index {
-                    if areas[index].columns.end != column + 1 {
-                        assert_ne!(areas[index].rows.start, row);
-                        return Err(());
-                    }
-                }
-                if row == 1 {
-                    width = column;
-                } else if width != column {
-                    return Err(());
-                }
+                },
             }
         }
+        if column == 0 {
+            // Each string must produce a valid token.
+            // https://github.com/w3c/csswg-drafts/issues/5110
+            return Err(());
+        }
+        if let Some(index) = current_area_index {
+            if self.areas[index].columns.end != column + 1 {
+                debug_assert_ne!(self.areas[index].rows.start, self.row);
+                return Err(());
+            }
+        }
+        if self.row == 1 {
+            self.width = column;
+        } else if self.width != column {
+            return Err(());
+        }
+
+        self.strings.push(simplified_string.into());
+        Ok(())
+    }
+
+    /// Return the parsed template areas.
+    pub fn finish(self) -> Result<TemplateAreas, ()> {
+        if self.strings.is_empty() {
+            return Err(());
+        }
         Ok(TemplateAreas {
-            areas: areas.into(),
-            strings: strings.into(),
-            width,
+            areas: self.areas.into(),
+            strings: self.strings.into(),
+            width: self.width,
         })
     }
 }
 
+impl TemplateAreas {
+    fn parse_internal(input: &mut Parser) -> Result<Self, ()> {
+        let mut parser = TemplateAreasParser::default();
+        while parser.try_parse_string(input).is_ok() {}
+        parser.finish()
+    }
+}
+
 impl Parse for TemplateAreas {
     fn parse<'i, 't>(
-        _context: &ParserContext,
+        _: &ParserContext,
         input: &mut Parser<'i, 't>,
     ) -> Result<Self, ParseError<'i>> {
-        let mut strings = vec![];
-        while let Ok(string) =
-            input.try_parse(|i| i.expect_string().map(|s| s.as_ref().to_owned().into()))
-        {
-            strings.push(string);
-        }
-
-        TemplateAreas::from_vec(strings)
+        Self::parse_internal(input)
             .map_err(|()| input.new_custom_error(StyleParseErrorKind::UnspecifiedError))
     }
 }
