@@ -26,8 +26,6 @@
 #include "nsContentUtils.h"  // for nsAutoScriptBlocker
 #include "nsDocShell.h"
 #include "nsLayoutUtils.h"
-#include "mozilla/TimelineConsumers.h"
-#include "mozilla/CompositeTimelineMarker.h"
 #include "mozilla/StartupTimeline.h"
 
 using namespace mozilla;
@@ -904,28 +902,6 @@ bool nsView::IsRoot() const {
   return mViewManager->GetRootView() == this;
 }
 
-nsRect nsView::GetBoundsInParentUnits() const {
-  nsView* parent = GetParent();
-  nsViewManager* VM = GetViewManager();
-  if (this != VM->GetRootView() || !parent) {
-    return mDimBounds;
-  }
-  int32_t ourAPD = VM->AppUnitsPerDevPixel();
-  int32_t parentAPD = parent->GetViewManager()->AppUnitsPerDevPixel();
-  return mDimBounds.ScaleToOtherAppUnitsRoundOut(ourAPD, parentAPD);
-}
-
-nsPoint nsView::ConvertFromParentCoords(nsPoint aPt) const {
-  const nsView* parent = GetParent();
-  if (parent) {
-    aPt = aPt.ScaleToOtherAppUnits(
-        parent->GetViewManager()->AppUnitsPerDevPixel(),
-        GetViewManager()->AppUnitsPerDevPixel());
-  }
-  aPt -= GetPosition();
-  return aPt;
-}
-
 static bool IsPopupWidget(nsIWidget* aWidget) {
   return aWidget->GetWindowType() == WindowType::Popup;
 }
@@ -1102,22 +1078,10 @@ void nsView::DidCompositeWindow(mozilla::layers::TransactionId aTransactionId,
   if (aCompositeStart == aCompositeEnd) {
     return;
   }
-
-  nsIDocShell* docShell = context->GetDocShell();
-
-  if (TimelineConsumers::HasConsumer(docShell)) {
-    TimelineConsumers::AddMarkerForDocShell(
-        docShell, MakeUnique<CompositeTimelineMarker>(
-                      aCompositeStart, MarkerTracingType::START));
-    TimelineConsumers::AddMarkerForDocShell(
-        docShell, MakeUnique<CompositeTimelineMarker>(aCompositeEnd,
-                                                      MarkerTracingType::END));
-  }
 }
 
 void nsView::RequestRepaint() {
-  PresShell* presShell = mViewManager->GetPresShell();
-  if (presShell) {
+  if (PresShell* presShell = mViewManager->GetPresShell()) {
     presShell->ScheduleViewManagerFlush();
   }
 }
