@@ -49,8 +49,6 @@ using namespace mozilla::layers;
 
 #undef DEBUG_MOUSE_LOCATION
 
-// Weakly held references to all of the view managers
-StaticAutoPtr<nsTArray<nsViewManager*>> nsViewManager::gViewManagers;
 uint32_t nsViewManager::gLastUserEventTime = 0;
 
 nsViewManager::nsViewManager()
@@ -60,14 +58,7 @@ nsViewManager::nsViewManager()
       mRefreshDisableCount(0),
       mPainting(false),
       mRecursiveRefreshPending(false),
-      mHasPendingWidgetGeometryChanges(false) {
-  if (gViewManagers == nullptr) {
-    // Create an array to hold a list of view managers
-    gViewManagers = new nsTArray<nsViewManager*>;
-  }
-
-  gViewManagers->AppendElement(this);
-}
+      mHasPendingWidgetGeometryChanges(false) {}
 
 nsViewManager::~nsViewManager() {
   if (mRootView) {
@@ -78,22 +69,6 @@ nsViewManager::~nsViewManager() {
 
   mRootViewManager = nullptr;
 
-  NS_ASSERTION(gViewManagers != nullptr, "About to use null gViewManagers");
-
-#ifdef DEBUG
-  bool removed =
-#endif
-      gViewManagers->RemoveElement(this);
-  NS_ASSERTION(
-      removed,
-      "Viewmanager instance was not in the global list of viewmanagers");
-
-  if (gViewManagers->IsEmpty()) {
-    // There aren't any more view managers so
-    // release the global array of view managers
-    gViewManagers = nullptr;
-  }
-
   MOZ_RELEASE_ASSERT(!mPresShell,
                      "Releasing nsViewManager without having called Destroy on "
                      "the PresShell!");
@@ -480,6 +455,7 @@ void nsViewManager::PostPendingUpdate() {
   nsViewManager* rootVM = RootViewManager();
   rootVM->mHasPendingWidgetGeometryChanges = true;
   if (rootVM->mPresShell) {
+    rootVM->mPresShell->SetNeedLayoutFlush();
     rootVM->mPresShell->ScheduleViewManagerFlush();
   }
 }
@@ -814,8 +790,7 @@ void nsViewManager::MoveViewTo(nsView* aView, nscoord aX, nscoord aY) {
   aView->SetPosition(aX, aY);
 }
 
-void nsViewManager::ResizeView(nsView* aView, const nsRect& aRect,
-                               bool aRepaintExposedAreaOnly) {
+void nsViewManager::ResizeView(nsView* aView, const nsRect& aRect) {
   NS_ASSERTION(aView->GetViewManager() == this, "wrong view manager");
 
   nsRect oldDimensions = aView->GetDimensions();
@@ -960,22 +935,33 @@ void nsViewManager::UpdateWidgetGeometry() {
   }
 }
 
+/* static */ void nsViewManager::CollectVMsForWillPaint(
+    nsView* aView, nsViewManager* aParentVM,
+    nsTArray<RefPtr<nsViewManager>>& aVMs) {
+  nsViewManager* vm = aView->GetViewManager();
+  if (vm != aParentVM) {
+    aVMs.AppendElement(vm);
+  }
+
+  for (nsView* child = aView->GetFirstChild(); child;
+       child = child->GetNextSibling()) {
+    CollectVMsForWillPaint(child, vm, aVMs);
+  }
+}
+
 void nsViewManager::CallWillPaintOnObservers() {
   MOZ_ASSERT(IsRootVM(), "Must be root VM for this to be called!");
 
-  if (NS_WARN_IF(!gViewManagers)) {
+  if (!mRootView) {
     return;
   }
 
-  uint32_t index;
-  for (index = 0; index < gViewManagers->Length(); index++) {
-    nsViewManager* vm = gViewManagers->ElementAt(index);
-    if (vm->RootViewManager() == this) {
-      // One of our kids.
-      if (vm->mRootView && vm->mRootView->IsEffectivelyVisible()) {
-        if (RefPtr<PresShell> presShell = vm->GetPresShell()) {
-          presShell->WillPaint();
-        }
+  AutoTArray<RefPtr<nsViewManager>, 2> VMs;
+  CollectVMsForWillPaint(mRootView, nullptr, VMs);
+  for (const auto& vm : VMs) {
+    if (vm->GetRootView() && vm->GetRootView()->IsEffectivelyVisible()) {
+      if (RefPtr<PresShell> presShell = vm->GetPresShell()) {
+        presShell->WillPaint();
       }
     }
   }
